/*
 * (C) Copyright 2011-2012 Nuxeo SA (http://nuxeo.com/) and contributors.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Contributors:
 *     Wojciech Sulejman
 *     Florent Guillaume
 */
package org.nuxeo.ecm.platform.signature.core.sign;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static org.nuxeo.ecm.platform.signature.api.sign.SignatureService.StatusWithBlob.SIGNED_CURRENT;
import static org.nuxeo.ecm.platform.signature.api.sign.SignatureService.StatusWithBlob.SIGNED_OTHER;
import static org.nuxeo.ecm.platform.signature.api.sign.SignatureService.StatusWithBlob.UNSIGNABLE;
import static org.nuxeo.ecm.platform.signature.api.sign.SignatureService.StatusWithBlob.UNSIGNED;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.nuxeo.common.utils.FileUtils;
import org.nuxeo.ecm.core.api.Blob;
import org.nuxeo.ecm.core.api.CoreSession;
import org.nuxeo.ecm.core.api.DocumentModel;
import org.nuxeo.ecm.core.api.impl.blob.FileBlob;
import org.nuxeo.ecm.core.api.impl.blob.StringBlob;
import org.nuxeo.ecm.core.test.CoreFeature;
import org.nuxeo.ecm.core.test.DefaultRepositoryInit;
import org.nuxeo.ecm.core.test.annotations.Granularity;
import org.nuxeo.ecm.core.test.annotations.RepositoryConfig;
import org.nuxeo.ecm.directory.Session;
import org.nuxeo.ecm.directory.api.DirectoryService;
import org.nuxeo.ecm.platform.signature.api.exception.AlreadySignedException;
import org.nuxeo.ecm.platform.signature.api.sign.SignatureService;
import org.nuxeo.ecm.platform.signature.api.sign.SignatureService.SigningDisposition;
import org.nuxeo.ecm.platform.signature.api.sign.SignatureService.StatusWithBlob;
import org.nuxeo.ecm.platform.signature.api.user.CUserService;
import org.nuxeo.ecm.platform.usermanager.UserManager;
import org.nuxeo.runtime.test.runner.Deploy;
import org.nuxeo.runtime.test.runner.Features;
import org.nuxeo.runtime.test.runner.FeaturesRunner;

import com.google.inject.Inject;
import com.lowagie.text.pdf.PdfReader;

@RunWith(FeaturesRunner.class)
@Features(CoreFeature.class)
@RepositoryConfig(user = "Administrator", init = DefaultRepositoryInit.class, cleanup = Granularity.METHOD)
@Deploy({ "org.nuxeo.ecm.core", //
        "org.nuxeo.ecm.core.api", //
        "org.nuxeo.runtime.management", //
        "org.nuxeo.ecm.directory", //
        "org.nuxeo.ecm.directory.sql", //
        "org.nuxeo.ecm.platform.usermanager", //
        "org.nuxeo.ecm.platform.usermanager.api", //
        "org.nuxeo.ecm.core.convert.api", //
        "org.nuxeo.ecm.core.convert", //
        "org.nuxeo.ecm.core.convert.plugins", //
        "org.nuxeo.ecm.platform.convert", //
        "org.nuxeo.ecm.platform.signature.core", //
        "org.nuxeo.ecm.platform.signature.core.test" })
public class SignatureServiceTest {

    @Inject
    protected CUserService cUserService;

    @Inject
    protected SignatureService signatureService;

    @Inject
    protected UserManager userManager;

    @Inject
    protected DirectoryService directoryService;

    @Inject
    protected CoreSession session;

    private static final String ORIGINAL_PDF = "pdf-tests/original.pdf";

    private static final String SIGNED_PDF = "pdf-tests/signed.pdf";

    private static final String HELLO_TXT = "pdf-tests/hello.txt";

    private static final String USER_KEY_PASSWORD = "abc";

    private static final String CERTIFICATE_DIRECTORY_NAME = "certificate";

    private static final String DEFAULT_USER_ID = "hSimpson";

    private static final String SECOND_USER_ID = "mSimpson";

    private File origPdfFile;

    private File signedPdfFile;

    private File helloTxtFile;

    private DocumentModel user;

    private DocumentModel user2;

    /**
     * Signing Prerequisite: a user with a certificate needs to be present
     */
    @Before
    public void setUp() throws Exception {

        DocumentModel userModel = userManager.getBareUserModel();
        userModel.setProperty("user", "username", DEFAULT_USER_ID);
        userModel.setProperty("user", "firstName", "Homer");
        userModel.setProperty("user", "lastName", "Simpson");
        userModel.setProperty("user", "email", "hsimpson@springfield.com");
        userModel.setPathInfo("/", DEFAULT_USER_ID);
        user = userManager.createUser(userModel);

        userModel = userManager.getBareUserModel();
        userModel.setProperty("user", "username", SECOND_USER_ID);
        userModel.setProperty("user", "firstName", "Marge");
        userModel.setProperty("user", "lastName", "Simpson");
        userModel.setProperty("user", "email", "msimpson@springfield.com");
        userModel.setPathInfo("/", SECOND_USER_ID);
        user2 = userManager.createUser(userModel);

        DocumentModel certificate = cUserService.createCertificate(user,
                USER_KEY_PASSWORD);
        assertNotNull(certificate);

        DocumentModel certificate2 = cUserService.createCertificate(user2,
                USER_KEY_PASSWORD);
        assertNotNull(certificate2);

        origPdfFile = FileUtils.getResourceFileFromContext(ORIGINAL_PDF);
        signedPdfFile = FileUtils.getResourceFileFromContext(SIGNED_PDF);
        helloTxtFile = FileUtils.getResourceFileFromContext(HELLO_TXT);
    }

    @After
    public void tearDown() throws Exception {

        // delete certificates associated with user ids
        Session sqlSession = directoryService.open(CERTIFICATE_DIRECTORY_NAME);
        sqlSession.deleteEntry(DEFAULT_USER_ID);
        sqlSession.deleteEntry(SECOND_USER_ID);
        sqlSession.close();

        // delete users
        userManager.deleteUser(DEFAULT_USER_ID);
        userManager.deleteUser(SECOND_USER_ID);
    }

    @Test
    public void testSignPDF() throws Exception {
        SignatureServiceImpl ssi = (SignatureServiceImpl) signatureService;

        // first user signs
        FileBlob origBlob = new FileBlob(origPdfFile);
        assertEquals(UNSIGNED, ssi.getSigningStatus(origBlob, user));
        Blob signedBlob = signatureService.signPDF(origBlob, user,
                USER_KEY_PASSWORD, "test reason");
        assertNotNull(signedBlob);
        assertEquals(SIGNED_CURRENT, ssi.getSigningStatus(signedBlob, user));
        assertEquals(SIGNED_OTHER, ssi.getSigningStatus(signedBlob, user2));

        // try for the same user to sign the certificate again
        Blob signedBlobTwice = null;
        try {
            signedBlobTwice = signatureService.signPDF(signedBlob, user,
                    USER_KEY_PASSWORD, "test reason");
            fail("Should raise AlreadySignedException");
        } catch (AlreadySignedException e) {
            // ok
        }

        // try for the second user to sign the certificate
        signedBlobTwice = signatureService.signPDF(signedBlob, user2,
                USER_KEY_PASSWORD, "test reason");
        assertNotNull(signedBlobTwice);
        assertEquals(SIGNED_CURRENT,
                ssi.getSigningStatus(signedBlobTwice, user));
        assertEquals(SIGNED_CURRENT,
                ssi.getSigningStatus(signedBlobTwice, user2));

        // test presence of multiple signatures
        List<String> names = getSignatureNames(signedBlobTwice);
        assertEquals(2, names.size());
        assertEquals(Arrays.asList("Signature2", "Signature1"), names);
    }

    protected List<String> getSignatureNames(Blob blob) throws IOException {
        PdfReader reader = new PdfReader(blob.getStream());
        try {
            @SuppressWarnings("unchecked")
            List<String> names = reader.getAcroFields().getSignatureNames();
            return names;
        } finally {
            reader.close();
        }
    }

    @Test
    public void testGetCertificates() throws Exception {
        SignatureServiceImpl ssi = (SignatureServiceImpl) signatureService;

        // sign the original PDF file
        Blob signedBlob = signatureService.signPDF(new FileBlob(origPdfFile),
                user, USER_KEY_PASSWORD, "test reason");
        assertNotNull(signedBlob);
        // verify there are certificates in the signed file
        List<X509Certificate> certificates = ssi.getCertificates(signedBlob);
        assertTrue(
                "There has to be at least 1 certificate in a signed document",
                certificates.size() > 0);
        assertTrue(certificates.get(0).getSubjectDN().toString().contains(
                "CN=Homer Simpson"));
    }

    @Test
    public void testGetSigningStatus() throws Exception {
        Serializable pdfBlob = new FileBlob(origPdfFile, "application/pdf");
        Serializable signedBlob = new FileBlob(signedPdfFile, "application/pdf");
        Serializable otherBlob = new StringBlob("foo",
                "application/octet-stream");
        DocumentModel doc = session.createDocumentModel("File");
        StatusWithBlob swb;

        swb = signatureService.getSigningStatus(doc, null);
        assertEquals(UNSIGNABLE, swb.status);
        assertNull(swb.path);
        assertNull(swb.blob);

        // unsigned PDF in main file
        doc.setPropertyValue("file:content", pdfBlob);
        swb = signatureService.getSigningStatus(doc, null);
        assertEquals(UNSIGNED, swb.status);
        assertEquals("file:content", swb.path);
        assertEquals(pdfBlob, swb.blob);

        // signed PDF in main file
        doc.setPropertyValue("file:content", signedBlob);
        swb = signatureService.getSigningStatus(doc, null);
        assertEquals(SIGNED_OTHER, swb.status);
        assertEquals("file:content", swb.path);
        assertEquals(signedBlob, swb.blob);

        // not PDF in main file
        doc.setPropertyValue("file:content", otherBlob);
        swb = signatureService.getSigningStatus(doc, null);
        assertEquals(UNSIGNED, swb.status);
        assertEquals("file:content", swb.path);
        assertEquals(otherBlob, swb.blob);

        // no files attached
        doc.setPropertyValue("file:content", null);
        doc.setPropertyValue("files:files", null);
        swb = signatureService.getSigningStatus(doc, null);
        assertEquals(UNSIGNABLE, swb.status);
        assertNull(swb.path);
        assertNull(swb.blob);
        doc.setPropertyValue("files:files",
                (Serializable) Collections.emptyList());
        swb = signatureService.getSigningStatus(doc, null);
        assertEquals(UNSIGNABLE, swb.status);
        assertNull(swb.path);
        assertNull(swb.blob);

        // unsigned PDF attached
        List<Map<String, Serializable>> fileList = new ArrayList<Map<String, Serializable>>();
        fileList.add(Collections.singletonMap("file", pdfBlob));
        doc.setPropertyValue("files:files", (Serializable) fileList);
        swb = signatureService.getSigningStatus(doc, null);
        assertEquals(UNSIGNABLE, swb.status);
        assertNull(swb.path);
        assertNull(swb.blob);

        // signed PDF attached second
        fileList.clear();
        fileList.add(Collections.singletonMap("file", pdfBlob));
        fileList.add(Collections.singletonMap("file", signedBlob));
        doc.setPropertyValue("files:files", (Serializable) fileList);
        swb = signatureService.getSigningStatus(doc, null);
        assertEquals(SIGNED_OTHER, swb.status);
        assertEquals("files:files/1/file", swb.path);
        assertEquals(signedBlob, swb.blob);

        // and PDF as main file
        doc.setPropertyValue("file:content", pdfBlob);
        swb = signatureService.getSigningStatus(doc, null);
        assertEquals(SIGNED_OTHER, swb.status);
        assertEquals("files:files/1/file", swb.path);
        assertEquals(signedBlob, swb.blob);
    }

    @Test
    public void testSignDocumentReplace() throws Exception {
        Blob txtBlob = new FileBlob(helloTxtFile, "text/plain", null,
                "foo.txt", null);
        DocumentModel doc = session.createDocumentModel("File");
        doc.setPropertyValue("file:content", (Serializable) txtBlob);

        Blob signedBlob = signatureService.signDocument(doc, user,
                USER_KEY_PASSWORD, "test", false, SigningDisposition.REPLACE,
                null);

        assertEquals("foo.pdf", signedBlob.getFilename());
        assertEquals(Arrays.asList("Signature1"), getSignatureNames(signedBlob));
        assertEquals(signedBlob, doc.getPropertyValue("file:content"));
        assertEquals(Collections.emptyList(),
                doc.getPropertyValue("files:files"));
    }

    @Test
    public void testSignDocumentAttach() throws Exception {
        Blob txtBlob = new FileBlob(helloTxtFile, "text/plain", null,
                "foo.txt", null);
        DocumentModel doc = session.createDocumentModel("File");
        doc.setPropertyValue("file:content", (Serializable) txtBlob);

        Blob signedBlob = signatureService.signDocument(doc, user,
                USER_KEY_PASSWORD, "test", false, SigningDisposition.ATTACH,
                null);

        assertEquals("foo.pdf", signedBlob.getFilename());
        assertEquals(Arrays.asList("Signature1"), getSignatureNames(signedBlob));
        assertEquals(txtBlob, doc.getPropertyValue("file:content"));
        @SuppressWarnings("unchecked")
        List<Map<String, Serializable>> files = (List<Map<String, Serializable>>) doc.getPropertyValue("files:files");
        assertEquals(1, files.size());
        assertEquals(signedBlob, files.get(0).get("file"));
    }

    @Test
    public void testSignDocumentArchive() throws Exception {
        Blob txtBlob = new FileBlob(helloTxtFile, "text/plain", null,
                "foo.txt", null);
        DocumentModel doc = session.createDocumentModel("File");
        doc.setPropertyValue("file:content", (Serializable) txtBlob);

        Blob signedBlob = signatureService.signDocument(doc, user,
                USER_KEY_PASSWORD, "test", false, SigningDisposition.ARCHIVE,
                "foo archive.txt");

        assertEquals("foo.pdf", signedBlob.getFilename());
        assertEquals(Arrays.asList("Signature1"), getSignatureNames(signedBlob));
        assertEquals(signedBlob, doc.getPropertyValue("file:content"));
        @SuppressWarnings("unchecked")
        List<Map<String, Serializable>> files = (List<Map<String, Serializable>>) doc.getPropertyValue("files:files");
        assertEquals(1, files.size());
        Blob archivedBlob = (Blob) files.get(0).get("file");
        assertEquals("text/plain", archivedBlob.getMimeType());
        assertEquals("foo archive.txt", archivedBlob.getFilename());
        assertEquals("foo archive.txt", files.get(0).get("filename"));
    }

    @Test
    public void testSignPDFDocumentReplace() throws Exception {
        Blob pdfBlob = new FileBlob(origPdfFile, "application/pdf", null,
                "foo.pdf", null);
        DocumentModel doc = session.createDocumentModel("File");
        doc.setPropertyValue("file:content", (Serializable) pdfBlob);

        Blob signedBlob = signatureService.signDocument(doc, user,
                USER_KEY_PASSWORD, "test", false, SigningDisposition.REPLACE,
                null);

        assertEquals("foo.pdf", signedBlob.getFilename());
        assertEquals(Arrays.asList("Signature1"), getSignatureNames(signedBlob));
        assertEquals(signedBlob, doc.getPropertyValue("file:content"));
        assertEquals(Collections.emptyList(),
                doc.getPropertyValue("files:files"));
    }

    @Test
    public void testSignPDFDocumentAttach() throws Exception {
        Blob pdfBlob = new FileBlob(origPdfFile, "application/pdf", null,
                "foo.pdf", null);
        DocumentModel doc = session.createDocumentModel("File");
        doc.setPropertyValue("file:content", (Serializable) pdfBlob);

        Blob signedBlob = signatureService.signDocument(doc, user,
                USER_KEY_PASSWORD, "test", false, SigningDisposition.ATTACH,
                null);

        assertEquals("foo.pdf", signedBlob.getFilename());
        assertEquals(Arrays.asList("Signature1"), getSignatureNames(signedBlob));
        assertEquals(pdfBlob, doc.getPropertyValue("file:content"));
        @SuppressWarnings("unchecked")
        List<Map<String, Serializable>> files = (List<Map<String, Serializable>>) doc.getPropertyValue("files:files");
        assertEquals(1, files.size());
        assertEquals(signedBlob, files.get(0).get("file"));
    }

    @Test
    public void testSignPDFDocumentArchive() throws Exception {
        Blob pdfBlob = new FileBlob(origPdfFile, "application/pdf", null,
                "foo.pdf", null);
        DocumentModel doc = session.createDocumentModel("File");
        doc.setPropertyValue("file:content", (Serializable) pdfBlob);

        Blob signedBlob = signatureService.signDocument(doc, user,
                USER_KEY_PASSWORD, "test", false, SigningDisposition.ARCHIVE,
                "foo archive.pdf");

        assertEquals("foo.pdf", signedBlob.getFilename());
        assertEquals(Arrays.asList("Signature1"), getSignatureNames(signedBlob));
        assertEquals(signedBlob, doc.getPropertyValue("file:content"));
        @SuppressWarnings("unchecked")
        List<Map<String, Serializable>> files = (List<Map<String, Serializable>>) doc.getPropertyValue("files:files");
        assertEquals(1, files.size());
        Blob archivedBlob = (Blob) files.get(0).get("file");
        assertEquals("application/pdf", archivedBlob.getMimeType());
        assertEquals("foo archive.pdf", archivedBlob.getFilename());
        assertEquals("foo archive.pdf", files.get(0).get("filename"));
    }

    @Test
    public void testResignDocument() throws Exception {
        Blob pdfBlob = new FileBlob(signedPdfFile, "application/pdf", null,
                "foo.pdf", null);
        assertEquals(Arrays.asList("Signature1"), getSignatureNames(pdfBlob));

        DocumentModel doc = session.createDocumentModel("File");
        doc.setPropertyValue("file:content", (Serializable) pdfBlob);

        Blob signedBlob = signatureService.signDocument(doc, user,
                USER_KEY_PASSWORD, "test", false, SigningDisposition.REPLACE,
                null);

        assertEquals("foo.pdf", signedBlob.getFilename());
        assertEquals(Arrays.asList("Signature2", "Signature1"),
                getSignatureNames(signedBlob));
    }

}