package org.nuxeo.ecm.platform.template.tests;

import java.io.File;
import java.util.List;

import org.junit.Test;
import org.nuxeo.common.utils.FileUtils;
import org.nuxeo.ecm.core.api.Blob;
import org.nuxeo.ecm.core.api.DocumentModel;
import org.nuxeo.ecm.core.api.impl.blob.FileBlob;
import org.nuxeo.ecm.core.event.EventService;
import org.nuxeo.ecm.core.storage.sql.SQLRepositoryTestCase;
import org.nuxeo.runtime.api.Framework;
import org.nuxeo.template.api.TemplateInput;
import org.nuxeo.template.api.adapters.TemplateBasedDocument;
import org.nuxeo.template.api.adapters.TemplateSourceDocument;

import static org.junit.Assert.*;

public class TestAdapters extends SQLRepositoryTestCase {

    @Override
    public void setUp() throws Exception {
        super.setUp();
        deployBundle("org.nuxeo.ecm.core.api");
        deployBundle("org.nuxeo.ecm.core");
        deployBundle("org.nuxeo.ecm.core.schema");
        deployBundle("org.nuxeo.ecm.core.event");
        deployBundle("org.nuxeo.template.manager.api");
        deployContrib("org.nuxeo.template.manager",
                "OSGI-INF/core-types-contrib.xml");
        deployContrib("org.nuxeo.template.manager",
                "OSGI-INF/life-cycle-contrib.xml");
        deployContrib("org.nuxeo.template.manager",
                "OSGI-INF/adapter-contrib.xml");
        openSession();
    }

    @Override
    public void tearDown() throws Exception {
        EventService eventService = Framework.getLocalService(EventService.class);
        eventService.waitForAsyncCompletion();
        closeSession();
        super.tearDown();
    }

    @Test
    public void testAdapters() throws Exception {

        // create a template doc
        DocumentModel root = session.getRootDocument();
        DocumentModel template = session.createDocumentModel(
                root.getPathAsString(), "template", "TemplateSource");
        template.setProperty("dublincore", "title", "MyTemplate");
        File file = FileUtils.getResourceFileFromContext("data/hello.docx");
        assertNotNull(file);
        Blob fileBlob = new FileBlob(file);
        fileBlob.setFilename("hello.docx");
        template.setProperty("file", "content", fileBlob);
        template = session.createDocument(template);

        // test the adapter
        TemplateSourceDocument templateSource = template.getAdapter(TemplateSourceDocument.class);
        assertNotNull(templateSource);

        Blob blob = templateSource.getTemplateBlob();
        assertNotNull(blob);
        assertEquals("hello.docx", blob.getFilename());

        // add fields
        TemplateInput input1 = new TemplateInput("field1", "Value1");
        input1.setDesciption("Some description");

        templateSource.addInput(input1);
        assertNotNull(templateSource.getAdaptedDoc().getPropertyValue(
                "tmpl:templateData"));
        assertEquals(1, templateSource.getParams().size());
        assertNotNull(templateSource.getParamsAsString());

        template = templateSource.save();

        session.save();

        // Test template based doc

        DocumentModel testDoc = session.createDocumentModel(
                root.getPathAsString(), "templatedDoc", "TemplateBasedFile");
        testDoc.setProperty("dublincore", "title", "MyTestDoc");
        testDoc = session.createDocument(testDoc);

        TemplateBasedDocument adapter = testDoc.getAdapter(TemplateBasedDocument.class);
        assertNotNull(adapter);

        // associated to template
        testDoc = adapter.setTemplate(templateSource.getAdaptedDoc(), true);
        List<String> templateNames = adapter.getTemplateNames();
        assertNotNull(templateNames);
        assertEquals(1, templateNames.size());
        assertEquals(templateSource.getName(), templateNames.get(0));

        testDoc = adapter.initializeFromTemplate(templateSource.getName(), true);

        // check fields
        List<TemplateInput> copiedParams = adapter.getParams(templateSource.getName());
        assertNotNull(copiedParams);
        assertEquals(1, copiedParams.size());
        assertEquals("field1", copiedParams.get(0).getName());
        assertEquals("Value1", copiedParams.get(0).getStringValue());

        // check update
        copiedParams.get(0).setStringValue("newValue");
        adapter.saveParams(templateSource.getName(), copiedParams, true);
        copiedParams = adapter.getParams(templateSource.getName());
        assertEquals("newValue", copiedParams.get(0).getStringValue());

        // create a second template doc
        DocumentModel template2 = session.createDocumentModel(
                root.getPathAsString(), "template2", "TemplateSource");
        template2.setProperty("dublincore", "title", "MyTemplate2");
        file = FileUtils.getResourceFileFromContext("data/testDoc.odt");
        assertNotNull(file);
        fileBlob = new FileBlob(file);
        fileBlob.setFilename("hello.odt");
        template2.setProperty("file", "content", fileBlob);
        template2 = session.createDocument(template2);

        // test the adapter
        TemplateSourceDocument templateSource2 = template2.getAdapter(TemplateSourceDocument.class);
        assertNotNull(templateSource2);

        // associated to template
        testDoc = adapter.setTemplate(templateSource2.getAdaptedDoc(), true);
        templateNames = adapter.getTemplateNames();
        assertNotNull(templateNames);
        assertEquals(2, templateNames.size());
        assertTrue(templateNames.contains(templateSource.getName()));
        assertTrue(templateNames.contains(templateSource2.getName()));

    }

}
