/*
 * (C) Copyright 2006-20012 Nuxeo SA (http://nuxeo.com/) and contributors.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Contributors:
 *     Razvan Caraghin
 *     Florent Guillaume
 *     Antoine Taillefer
 */

package org.nuxeo.ecm.webapp.versioning;

import org.jboss.seam.annotations.Create;
import org.nuxeo.ecm.core.api.ClientException;
import org.nuxeo.ecm.core.api.DocumentModel;
import org.nuxeo.ecm.core.api.VersionModel;
import org.nuxeo.ecm.platform.query.api.PageSelections;

/**
 * Exposes the actions that can be taken related to versioning and documents.
 *
 * @author Razvan Caraghin
 * @author Florent Guillaume
 */
public interface VersionedActions {

    /**
     * Factory accessor for currentDocument versionList.
     *
     * @return the selected version list as a {@link PageSelections<VersionModel>}
     */
    PageSelections<VersionModel> getVersionList() throws ClientException;

    /**
     * Retrieves the versions for the current document.
     */
    void retrieveVersions() throws ClientException;

    /**
     * Restored the document to the selected version. If there is no selected
     * version it does nothing.
     *
     * @return the page that needs to be displayed next
     */
    String restoreToVersion(VersionModel selectedVersion)
            throws ClientException;

    /**
     * Restores the version which id is returned by
     * {@link #getSelectedVersionId()}.
     *
     * @return the view id
     * @throws ClientException if cannot fetch the version
     *
     * @since 5.6
     */
    String restoreToVersion() throws ClientException;

    /**
     * Security check to enable or disable the restore button.
     *
     * @return permission check result
     */
    boolean getCanRestore() throws ClientException;

    /**
     * Tells if the current selected document is checked out or not.
     */
    String getCheckedOut() throws ClientException;

    /**
     * Changes the checked-out string.
     *
     * @param checkedOut
     */
    void setCheckedOut(String checkedOut);

    /**
     * Checks the document out.
     *
     * @return the next page
     */
    @SuppressWarnings({ "NonBooleanMethodNameMayNotStartWithQuestion" })
    String checkOut() throws ClientException;

    /**
     * Checks the selected document in, with the selected version.
     */
    String checkIn() throws ClientException;

    @Create
    void initialize();

    /**
     * When the user selects/changes other documents then we nullify the list of
     * versions associated with the document so that the factory method gets
     * called when the list is used.
     * <p>
     * This way we achieve lazy loading of data from backend - only when its
     * needed and not loading it when the event is fired.
     *
     */
    void resetVersions();

    /**
     * View an older version of the document.
     */
    String viewArchivedVersion(VersionModel selectedVersion)
            throws ClientException;

    /**
     * Navigates to the version which id is returned by
     * {@link #getSelectedVersionId()}.
     *
     * @return the view id
     * @throws ClientException if cannot fetch the version
     *
     * @since 5.6
     */
    String viewArchivedVersion() throws ClientException;

    DocumentModel getSourceDocument() throws ClientException;

    DocumentModel getSourceDocument(DocumentModel document)
            throws ClientException;

    /**
     * Check if a version can be removed. It won't be possible if a proxy is
     * pointing to it.
     */
    boolean canRemoveArchivedVersion(VersionModel selectedVersion);

    /**
     * Check if the currently selected versions can be removed. It won't be
     * possible if a proxy is pointing to one of them.
     *
     * @return true if can remove selected archived versions
     * @throws ClientException if at least one of the selected versions cannot
     *             be retrieved from the repository
     * @since 5.6
     */
    boolean getCanRemoveSelectedArchivedVersions() throws ClientException;

    /**
     * Remove an archived version.
     *
     * @param selectedVersion the version model to remove
     */
    String removeArchivedVersion(VersionModel selectedVersion)
            throws ClientException;

    /**
     * Remove currently selected archived versions.
     *
     * @since 5.6
     */
    String removeSelectedArchivedVersions() throws ClientException;

    /**
     * Gets currently selected version id.
     *
     * @since 5.6
     */
    String getSelectedVersionId();

    /**
     * Sets currently selected version id.
     *
     * @since 5.6
     */
    void setSelectedVersionId(String selectedVersionId);
}
